# SPDX-FileCopyrightText: 2025 Maurice Debray <maurice.debray@dgnum.eu>
# SPDX-FileCopyrightText: 2025 Tom Hubrecht <tom.hubrecht@dgnum.eu>
# SPDX-FileCopyrightText: 2025 Clans contributers
#
# SPDX-License-Identifier: MIT

/*
  Copyright 2023 Clan contributers, Tom-Hubrecht, Maurice Debray

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to
  deal in the Software without restriction, including without limitation the
  rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  sell copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in
  all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  DEALINGS IN THE SOFTWARE.
*/

{
  config,
  lib,
  pkgs,
  options,
  ...
}:

let
  inherit (lib)
    boolToString
    concatStringsSep
    escape
    filterAttrs
    floatToString
    genAttrs
    genList
    getVersion
    hasPrefix
    isBool
    isDerivation
    isFloat
    isInt
    isList
    isString
    mapAttrsToList
    mkAfter
    mkDefault
    mkEnableOption
    mkIf
    mkOption
    optional
    optionalString
    recursiveUpdate
    stringLength
    strings
    toPretty
    versionAtLeast
    ;

  inherit (lib.types)
    bool
    either
    ints
    listOf
    nullOr
    package
    path
    str
    ;

  toString' = n: (s: if stringLength s == 1 then "0${s}" else s) (builtins.toString n);

  runnerPath = lib.makeBinPath cfg.dependencies;

  cfg = config.services.forgejo-multiuser-nix-runners;

  storePaths = "nix-store --query -R ${builtins.toString cfg.dependencies} ${builtins.storePath pkgs.path}";

  nixConf =
    let
      isNixAtLeast = versionAtLeast (getVersion config.nix.package.out);
      mkValueString =
        v:
        if v == null then
          ""
        else if isInt v then
          toString v
        else if isBool v then
          boolToString v
        else if isFloat v then
          floatToString v
        else if isList v then
          toString v
        else if isDerivation v then
          toString v
        else if builtins.isPath v then
          toString v
        else if isString v then
          v
        else if strings.isConvertibleWithToString v then
          toString v
        else
          abort "The nix conf value: ${toPretty { } v} can not be encoded";

      mkKeyValue = k: v: "${escape [ "=" ] k} = ${mkValueString v}";

      mkKeyValuePairs = attrs: concatStringsSep "\n" (mapAttrsToList mkKeyValue attrs);

      isExtra = key: hasPrefix "extra-" key;

    in
    pkgs.writeTextFile rec {
      name = "nix.conf";
      destination = "/" + name;
      # workaround for https://github.com/NixOS/nix/issues/9487
      # extra-* settings must come after their non-extra counterpart
      text = ''
        ${mkKeyValuePairs (filterAttrs (key: _value: !(isExtra key)) cfg.nixConfig)}
        ${mkKeyValuePairs (filterAttrs (key: _value: isExtra key) cfg.nixConfig)}
      '';
      checkPhase =
        if pkgs.stdenv.hostPlatform != pkgs.stdenv.buildPlatform then
          ''
            echo "Ignoring validation for cross-compilation"
          ''
        else
          let
            showCommand = if isNixAtLeast "2.20pre" then "config show" else "show-config";
          in
          ''
            echo "Validating generated nix.conf"
            ln -s $out ./nix.conf
            set -e
            set +o pipefail
            NIX_CONF_DIR=$PWD \
              ${pkgs.nix}/bin/nix ${showCommand} ${optionalString (isNixAtLeast "2.3pre") "--no-net"} \
                ${optionalString (isNixAtLeast "2.4pre") "--option experimental-features nix-command"} \
              |& sed -e 's/^warning:/error:/' \
              | (! grep '^error:')
            set -o pipefail
          '';
    };

  mkRunnerService = _: {
    after = [ "forgejo-multiuser-nix-runner-image.service" ];
    requires = [ "forgejo-multiuser-nix-runner-image.service" ];

    # TODO: systemd confinment
    serviceConfig = {
      # Hardening (may overlap with DynamicUser=)
      # The following options are only for optimizing output of systemd-analyze
      AmbientCapabilities = "";
      CapabilityBoundingSet = "";
      # ProtectClock= adds DeviceAllow=char-rtc r
      DeviceAllow = "";
      NoNewPrivileges = true;
      PrivateDevices = true;
      PrivateMounts = true;
      PrivateTmp = true;
      PrivateUsers = true;
      ProtectClock = true;
      ProtectControlGroups = true;
      ProtectHome = true;
      ProtectHostname = true;
      ProtectKernelLogs = true;
      ProtectKernelModules = true;
      ProtectKernelTunables = true;
      ProtectSystem = "strict";
      RemoveIPC = true;
      RestrictNamespaces = true;
      RestrictRealtime = true;
      RestrictSUIDSGID = true;
      UMask = "0066";
      ProtectProc = "invisible";
      SystemCallFilter = [
        "~@clock"
        "~@cpu-emulation"
        "~@module"
        "~@mount"
        "~@obsolete"
        "~@raw-io"
        "~@reboot"
        "~@swap"
        # needed by go?
        #"~@resources"
        "~@privileged"
        "~capset"
        "~setdomainname"
        "~sethostname"
      ];
      RestrictAddressFamilies = [
        "AF_INET"
        "AF_INET6"
        "AF_UNIX"
        "AF_NETLINK"
      ];

      # Needs network access
      PrivateNetwork = false;
      # Cannot be true due to Node
      MemoryDenyWriteExecute = false;

      # The more restrictive "pid" option makes `nix` commands in CI emit
      # "GC Warning: Couldn't read /proc/stat"
      # You may want to set this to "pid" if not using `nix` commands
      ProcSubset = "all";
      # Coverage programs for compiled code such as `cargo-tarpaulin` disable
      # ASLR (address space layout randomization) which requires the
      # `personality` syscall
      # You may want to set this to `true` if not using coverage tooling on
      # compiled code
      LockPersonality = false;

      ReadWritePaths = [ cfg.storePath ];

      DynamicUser = true;
    };
  };
in
{
  options.services.forgejo-multiuser-nix-runners = {
    enable = mkEnableOption "Forgero Actions Runners based on Nix";

    nixConfig = options.nix.settings // {
      default = {
        accept-flake-config = true;
        experimental-features = [
          "flakes"
          "nix-command"
        ];
      };
      description = ''
        Configuration for the nix version used inside the runners. See
        <https://nixos.org/manual/nix/stable/command-ref/conf-file.html> or
        {manpage}`nix.conf(5)` for available options.
        The value declared here will be translated directly to the key-value pairs Nix expects.
      '';
    };

    nbRunners = mkOption {
      type = ints.positive;
      description = "Number of runners to launch.";
      default = 4;
    };

    storePath = mkOption {
      type = str;
      description = "Path to the nix directory";
      default = "/";
    };

    dependencies = mkOption {
      type = listOf package;
      description = "List of packages to be made available directly to the runners.";
    };

    hostPlatform = mkOption {
      type = nullOr str;
      default = null;
      description = ''
        The host platform that will be running the containers.
        When set, a new label `nix-$hostPlatform` will be added, can be useful for compiling native systems.
      '';
    };

    names = mkOption {
      type = listOf str;
      description = ''
        List of names for the runners, it will create labels pointing to the correct docker image.
      '';
    };

    labels = mkOption {
      type = listOf str;
      description = ''
        Additional labels for the containers.
      '';
    };

    enableKvm = mkOption {
      type = bool;
      default = true;
      description = ''
        Whether to pass /dev/kvm as a device to the container.
      '';
    };

    containerOptions = mkOption {
      type = listOf str;
      description = "Extra options to pass to the containers.";
      default = [ ];
    };

    url = mkOption {
      type = str;
      example = "https://forge.example.com";
      description = ''
        Base URL of your Forgejo instance.
      '';
    };

    tokenFile = mkOption {
      type = nullOr (either str path);
      default = null;
      description = ''
        Path to an environment file, containing the `TOKEN` environment
        variable, that holds a token to register at the configured
        Forgejo instance.
      '';
    };

    settings = mkOption {
      inherit (pkgs.formats.yaml { }) type;
      default = { };
      description = ''
        Settings that will be applied to all runners.
      '';
    };
  };

  config = mkIf cfg.enable {
    services.forgejo-multiuser-nix-runners = {
      dependencies = [
        pkgs.bash
        pkgs.coreutils
        pkgs.curl
        pkgs.findutils
        pkgs.forgejo-cli
        pkgs.gawk
        pkgs.git
        pkgs.gnugrep
        pkgs.jq
        pkgs.nix
        pkgs.nodejs
        pkgs.openssh
      ];

      nixConfig = {
        allowed-users = [ "nixuser" ];
        trusted-users = [ "nixrootuser" ];
        trusted-public-keys = [ "cache.nixos.org-1:6NCHdD59X431o0gWypbMrAURkbJ16ZPMQFGspcDShjY=" ];
        substituters = mkAfter [ "https://cache.nixos.org/" ];
        system-features = mkDefault config.nix.settings.system-features;
      };

      names = [ "nix" ] ++ (optional (cfg.hostPlatform != null) "nix-${cfg.hostPlatform}");
      labels = builtins.map (name: "${name}:docker://forgejo-multiuser-nix-runner") cfg.names;
    };

    systemd = {
      sockets = {
        forgejo-multiuser-nix-daemon = mkIf (cfg.storePath != "/") {
          unitConfig = {
            RequiresMountsFor = "${cfg.storePath}/nix/store";
          };

          socketConfig = {
            ListenStream = "${cfg.storePath}/nix/var/nix/daemon-socket/socket";
          };
        };

      };

      tmpfiles.settings."forgejo-multiuser-runners" = {
        ${cfg.storePath}.d = {
          mode = "0755";
          group = "nixrootuser";
          user = "nixrootuser";
        };
        "${cfg.storePath}/nix".d = {
          mode = "0755";
          group = "nixrootuser";
          user = "nixrootuser";
        };
        "${cfg.storePath}/nix/var".d = {
          mode = "0755";
          group = "nixrootuser";
          user = "nixrootuser";
        };
        "${cfg.storePath}/nix/var/nix".d = {
          mode = "0755";
          group = "nixrootuser";
          user = "nixrootuser";
        };
        "${cfg.storePath}/nix/var/nix/daemon-socket".d = {
          mode = "0755";
          group = "nixrootuser";
          user = "nixrootuser";
        };
      };
      services =
        (genAttrs (genList (n: "gitea-runner-multiuser${toString' n}") cfg.nbRunners) mkRunnerService)
        // {
          forgejo-multiuser-nix-store = mkIf (cfg.storePath != "/") {
            wantedBy = [ "multi-user.target" ];
            path = [
              pkgs.curl
              pkgs.nix
            ];
            script = ''
              set -o pipefail
              nix-store --export $(${storePaths}) | NIX_REMOTE="unix://${cfg.storePath}/nix/var/nix/daemon-socket/socket" nix-store --import
            ''; # TODO: benchmark this (I believe it is very slow)

            serviceConfig = {
              User = "nixrootuser";
              RuntimeDirectory = "forgejo-multiuser-nix-store";
              WorkingDirectory = "/run/forgejo-multiuser-nix-store";
              Type = "oneshot";
              RemainAfterExit = true;
              UMask = "0066";
              ReadWritePaths = [
                cfg.storePath
              ];
            };
          };

          forgejo-multiuser-nix-daemon = mkIf (cfg.storePath != "/") {
            wantedBy = [ "multi-user.target" ];
            serviceConfig = {
              ExecStart = "${pkgs.nix}/bin/nix-daemon --store ${cfg.storePath}";
              User = "nixrootuser";
              RuntimeDirectory = "forgejo-multiuser-nix-store";
              WorkingDirectory = "/run/forgejo-multiuser-nix-store";
              KillMode = "process";
              LimitNOFILE = 1048576;
              TasksMax = 1048576;
              CPUSchedulingPolicy = "other";
              Delegate = "yes";
              IOSchedulingClass = "best-effort";
              IOSchedulingPriority = 4;
              ReadWritePaths = [
                cfg.storePath
              ];

              #FIXME Harden
            };

            unitConfig = {
              Requires = [ "forgejo-multiuser-nix-daemon.socket" ]; # If we don't have the socket, this will fail
              RequiresMountsFor = [
                "${cfg.storePath}/nix/store"
                "${cfg.storePath}/nix/var"
                "${cfg.storePath}/nix/var/nix/db"
              ];
            };

            environment = {
              CURL_CA_BUNDLE = "${pkgs.cacert}/etc/ssl/certs/ca-certificates.crt";
              NIX_CONF_DIR = nixConf;
            };
            path = [
              pkgs.nix
              pkgs.util-linux
              pkgs.openssh
            ];
          };

          forgejo-multiuser-nix-runner-image = {
            wantedBy = [ "multi-user.target" ];
            after = [
              "podman.service"
              "forgejo-multiuser-nix-store.service"
            ];
            requires = [
              "podman.service"
              "forgejo-multiuser-nix-store.service"
            ];
            path = [
              config.virtualisation.podman.package
              pkgs.gnutar
              pkgs.shadow
              pkgs.getent
            ];

            # we also include etc here because the cleanup job also wants the nixuser to be present
            script = ''
              set -eux -o pipefail

              mkdir -p etc/nix

              echo 'root:x:0:0:System administrator:/root:/bin/bash' > etc/passwd
              echo 'root:x:0:' > etc/group

              # Create an unpriveleged user that we can use also without the run-as-user.sh script
              groupid=$(cut -d: -f3 < <(getent group nixuser))
              userid=$(cut -d: -f3 < <(getent passwd nixuser))
              groupadd --prefix $(pwd) --gid "$groupid" nixuser
              emptypassword='$6$1ero.LwbisiU.h3D$GGmnmECbPotJoPQ5eoSTD6tTjKnSWZcjHoVTkxFLZP17W9hRi/XkmCiAMOfWruUwy8gMjINrBMNODc7cYEo4K.'
              useradd --prefix $(pwd) -p "$emptypassword" -m -d /tmp -u "$userid" -g "$groupid" -G nixuser nixuser

              cp ${nixConf}/nix.conf etc/nix/nix.conf

              cat <<NSSWITCH > etc/nsswitch.conf
              passwd:    files mymachines systemd
              group:     files mymachines systemd
              shadow:    files

              hosts:     files mymachines dns myhostname
              networks:  files

              ethers:    files
              services:  files
              protocols: files
              rpc:       files
              NSSWITCH

              # Link usr/bin/env
              mkdir -p usr/bin
              ln -s ${pkgs.coreutils}/bin/env usr/bin/env

              # Link bin/sh
              mkdir -p bin
              ln -s ${pkgs.bash}/bin/sh bin/sh


              # list the content as it will be imported into the container
              tar -cv . | tar -tvf -
              tar -cv . | podman import - forgejo-multiuser-nix-runner
            '';

            serviceConfig = {
              RuntimeDirectory = "forgejo-multiuser-nix-runner-image";
              WorkingDirectory = "/run/forgejo-multiuser-nix-runner-image";
              Type = "oneshot";
              RemainAfterExit = true;
            };
          };
        };
    };

    users = {
      users.nixuser = {
        group = "nixuser";
        description = "Used for running nix ci jobs";
        home = "/var/empty";
        isSystemUser = true;
      };
      users.nixrootuser = {
        group = "nixrootuser";
        description = "Used for running nix daemon for ci jobs";
        home = cfg.storePath;
        isSystemUser = true;
      };
      groups.nixuser = { };
      groups.nixrootuser = { };
    };

    services.gitea-actions-runner = {
      package = lib.mkForce pkgs.forgejo-actions-runner;

      instances = genAttrs (genList (n: "multiuser${toString' n}") cfg.nbRunners) (_: {
        enable = true;
        name = "multiuser-nix-runner";

        inherit (cfg) labels url tokenFile;

        settings = recursiveUpdate cfg.settings {
          container = {
            options = builtins.toString (
              [
                "-e PAGER=cat"
                "-e PATH=${runnerPath}"
                "-e NIX_PATH=nixpkgs=${pkgs.path}"
                "-e SSL_CERT_FILE=/etc/ssl/certs/ca-bundle.crt"
                "-v ${cfg.storePath}/nix:/nix:ro"
                "-v ${pkgs.cacert}/etc/ssl:/etc/ssl:ro"
                "--user nixuser"
              ]
              ++ (optional cfg.enableKvm "--device=/dev/kvm")
              ++ cfg.containerOptions
            );

            # the default network that also respects our dns server settings
            network = "host";
            valid_volumes = [
              "${cfg.storePath}/nix"
              "${pkgs.bash}/bin"
              "${pkgs.cacert}/etc/ssl"
            ];
          };
        };
      });
    };
  };
}
