# SPDX-FileCopyrightText: 2025 Tom Hubrecht <tom.hubrecht@dgnum.eu>
#
# SPDX-License-Identifier: EUPL-1.2

{
  lib,
  stdenv,
  fetchFromGitHub,

  # Ruby dependencies
  bundlerEnv,
  ruby_3_4,

  # JS dependencies
  fetchYarnDeps,
  fixup-yarn-lock,
  nodejs,
  yarn,

  # Update script dependencies
  writeShellApplication,
  bundix,
  curl,
  git,
  jq,
  nixfmt-rfc-style,
  nurl,
  prefetch-yarn-deps,
}:

let
  meta' = builtins.fromJSON (builtins.readFile ./deps.json);

  rubyEnv = bundlerEnv {
    name = "docuseal-gems";
    gemdir = ./rubyEnv;
    ruby = ruby_3_4;
  };
in

stdenv.mkDerivation (finalAttrs: {
  pname = "docuseal";
  inherit (meta') version;

  src = fetchFromGitHub {
    owner = "docusealco";
    repo = "docuseal";
    tag = finalAttrs.version;
    inherit (meta') hash;
  };

  patches = [ ];

  frontend = stdenv.mkDerivation {
    pname = "${finalAttrs.pname}-frontend";
    inherit (finalAttrs) patches src version;

    offlineCache = fetchYarnDeps {
      yarnLock = "${finalAttrs.src}/yarn.lock";
      inherit (meta'.yarn) hash;
    };

    buildInputs = [ rubyEnv ];
    nativeBuildInputs = [
      fixup-yarn-lock
      nodejs
      yarn
      rubyEnv.wrappedRuby
    ];

    configurePhase = ''
      runHook preConfigure

      export HOME=$(mktemp -d)
      yarn config --offline set yarn-offline-mirror $offlineCache
      cp --no-preserve=owner,mode ${./yarn.lock} ${./rubyEnv/Gemfile} ${./rubyEnv/Gemfile.lock} .
      fixup-yarn-lock yarn.lock

      yarn install --offline --frozen-lockfile --ignore-platform --ignore-scripts --no-progress --non-interactive

      patchShebangs node_modules/
      patchShebangs bin/

      runHook postConfigure
    '';

    buildPhase = ''
      runHook preBuild

      bin/rake assets:precompile
      bin/rake shakapacker:compile

      runHook postBuild
    '';

    installPhase = ''
      runHook preInstall

      mv public/packs $out

      runHook postInstall
    '';
  };

  propagatedBuildInputs = [ rubyEnv.wrappedRuby ];

  buildPhase = ''
    runHook preBuild

    ln -s $frontend public/packs

    rm -rf log tmp

    ln -s /var/log/docuseal log
    ln -s /var/cache/docuseal tmp

    cp --no-preserve=owner,mode ${./rubyEnv/Gemfile} ${./rubyEnv/Gemfile.lock} .

    patchShebangs bin/

    runHook postBuild
  '';

  installPhase = ''
    mkdir $out && cp -R . $out
  '';

  passthru.updateScript = writeShellApplication {
    name = "update";
    runtimeInputs = [
      bundix
      curl
      git
      jq
      nixfmt-rfc-style
      nurl
      prefetch-yarn-deps
      ruby_3_4
      yarn
    ];
    text = builtins.readFile ./update.sh;
  };

  meta = {
    description = "Open source DocuSign alternative. Create, fill, and sign digital documents";
    homepage = "https://github.com/docusealco/docuseal";
    license = lib.licenses.agpl3Only;
    mainProgram = "docuseal";
    platforms = lib.platforms.all;
  };
})
